@@ -2,7 +2,7 @@
 
 Copyright (C) 1998 Fabrice POPINEAU.
 
-Time-stamp: <99/01/20 13:46:58 popineau>
+Time-stamp: <99/05/10 17:59:51 popineau>
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
@@ -28,12 +28,15 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 #include <kpathsea/kpathsea.h>
 extern KPSEDLL char* kpathsea_version_string;
 
+#include "fileutils.h"
 #include "stackenv.h"
 #include "variables.h"
 #include "fmtutil.h"
 
 #ifdef _WIN32
 #define isatty _isatty
+#else
+#include <glob.h>
 #endif
 
 extern string fmtutil_version_string;
@@ -46,6 +49,8 @@ static string texmfmain;
 static string tempenv;
 static string cwd;
 
+static string mode, bdpi;
+
 /* fmtutil.cnf */
 
 char tmpdir[PATH_MAX]; /* Working temporary directory */
@@ -55,6 +60,7 @@ FILE *fnul = NULL;
 static boolean downcase_names;
 char cmd_buf[1024];			/* Hope it is enough for handling for command */
 char empty_str[] = "(empty)";
+char cmd_sys[1024];		/* system() calls */
 
 static int errstatus;
 
@@ -67,6 +73,10 @@ void usage(void);
 
 static struct option long_options [] = {
     { "all",               0, 0, 0},
+    { "quiet",             0, 0, 0},
+    { "test",              0, 0, 0},
+    { "force",             0, 0, 0},
+    { "dolinks",           0, 0, 0},
     { "missing",           0, 0, 0},
     { "byfmt",             1, 0, 0},
     { "byhyphen",          1, 0, 0},
@@ -85,9 +95,10 @@ enum {
 #define MAX_FMTS 128
 
 typedef struct fmtdesc {
-  string format;
-  string engine;
-  string hyphenation;
+  string format;		/* e.g.: latex.fmt */
+  string engine;		/* e.g.: tex */
+  string progname;		/* e.g.: latex */
+  string hyphenation;		/* e.g.: - */
   string texargs;
   string inifile;
 } fmtdesc;
@@ -95,6 +106,10 @@ typedef struct fmtdesc {
 static fmtdesc fmts[MAX_FMTS];
 static int nb_formats = 0;
 
+string unrmable_engines[] = {
+  "tex", "pdftex", "etex", "pdfetex", "omega", NULL
+};
+
 /*
   First  argument takes progname;
   Other arguments take no arguments.
@@ -103,7 +118,11 @@ static string usage_msg[MAX_LINES] = {
 	"Usage: fmtutil [option] ... cmd [argument]\n\n",
 	"Valid options:\n",
 	"  --cnffile file\n",
-	"  --fmtdir directory\n\n",
+	"  --fmtdir directory\n",
+	"  --quiet                    no output except error messages\n",
+	"  --test                     only print what would be done\n",
+	"  --dolinks                  link engine to format\n\n",
+	"  --force                    force links even if target exists\n\n",
 	"Valid commands:\n",
 	"  --all                      recreate all format files\n",
 	"  --missing                  create all missing format files\n",
@@ -114,13 +133,28 @@ static string usage_msg[MAX_LINES] = {
 };
 
 string progname = NULL;
-static string fmtutil_version_string = "Revision: 0.1";
+static string fmtutil_version_string = "Revision: 0.2";
 
 static string cnf_default = "fmtutil.cnf";
 static string cnf_file = NULL;
 static string arg = NULL;
 static string destdir = NULL;
 static int cmd= NONE;
+static boolean quiet = false;
+static boolean norebuild = false;
+static boolean dolinks = false;
+static boolean force = false;
+
+#define MAX_LINKS 128
+
+struct _links {
+  string src;
+  string dst;
+} to_link[MAX_LINKS];
+int nb_links = 0;
+
+void link_formats();
+
 /*
   First part: fmtutil.opt
   */
@@ -144,6 +178,18 @@ int fmtutil_opt(int argc, char *argv[])
     if (ARGUMENT_IS ("debug")) {
       kpathsea_debug |= atoi (optarg);
     }
+    else if (ARGUMENT_IS ("quiet")) {
+      quiet = true;
+    }
+    else if (ARGUMENT_IS ("test")) {
+      norebuild = true;
+    }
+    else if (ARGUMENT_IS ("dolinks")) {
+      dolinks = true;
+    }
+    else if (ARGUMENT_IS ("force")) {
+      force = true;
+    }
     else if (ARGUMENT_IS ("help")) {
       usage();
       exit(0);
@@ -180,19 +226,17 @@ int fmtutil_opt(int argc, char *argv[])
     }
   }
 
-  fprintf(stderr, "CMD = %d\n", cmd);
-
   /* shifting options from argv[] list */
   for (i = 1; optind < argc; i++, optind++)
     argv[i] = argv[optind];
   argv[i] = NULL;
 
   argc = i;
+#if 0
   fprintf(stderr, "New args [%d] : ", argc);
   for (i = 0; i < argc; i++)
     fprintf(stderr, "%s ", argv[i]);
   fprintf(stderr, "\n");
-#if 0
   if (argc < program->arg_min) {
     fprintf (stderr, "%s: Missing argument(s).\nTry `%s --help' for more information.\n", progname, kpse_program_name);
     exit(1);
@@ -227,12 +271,31 @@ boolean parse_line(string line)
   if (!texargs || !*texargs) 
     return false;
   for (p = texargs+strlen(texargs) - 1; 
-       !isspace(*p) && (*p != '*') && (p>texargs); p--);
-  inifile = p;
+       !isspace(*p) && (*p != '*') && (p>=texargs); p--);
+  inifile = ++p;
+
+#if 0
+  if (strcmp(inifile, "pdflatex.ini")== 0) {
+        kpathsea_debug=-1; 
+  }
+#endif
+  kpse_reset_program_name(format);
+#if 0
+  /* kpathsea 3.3beta5 should correct this */
+  kpse_format_info[kpse_tex_format].cnf_path = kpse_cnf_get("TEXINPUTS");
+#endif
   if (!kpse_find_file(inifile, kpse_tex_format, false)) {
     fprintf(stderr, "%s: ini file %s not found\n", cnf_file, inifile);
     return false;
   }
+  kpse_reset_program_name("fmtutil");
+#if 0
+  /* kpathsea 3.3beta5 should correct this */
+  kpse_format_info[kpse_tex_format].cnf_path = kpse_cnf_get("TEXINPUTS");
+  kpathsea_debug=0;
+#endif
+  fmts[nb_formats].progname = xstrdup(format);
+
   if (STREQ(engine, "etex")
       || STREQ(engine, "pdfetex")) {
     fmts[nb_formats].format = concat(format, ".efmt");
@@ -272,12 +335,12 @@ void read_fmtutilcnf()
       free(line);
     }
     fclose(f);
-    {
+    if (KPSE_DEBUG_P(FMTUTIL_DEBUG)) {
       int i;
       for (i = 0; i < nb_formats; i++) {
-	fprintf(stderr, "Line %d: fmt = %s engine = %s hyph = %s tex = %s ini = %s\n",
-		fmts[i].format , fmts[i].engine, fmts[i].hyphenation, fmts[i].texargs,
-		fmts[i].inifile);
+	fprintf(stderr, "Line %d: fmt = %s engine = %s progname = %s hyph = %s tex = %s ini = %s\n",
+		i, fmts[i].format, fmts[i].engine, fmts[i].progname,
+		fmts[i].hyphenation, fmts[i].texargs, fmts[i].inifile);
       }
     }
 
@@ -316,7 +379,6 @@ void usage()
 
 int main(int argc, char* argv[])
 {
-  string mode, bdpi;
   string texinputs;
 
 #if defined(WIN32)
@@ -351,7 +413,7 @@ int main(int argc, char* argv[])
 
   /* if no cnf_file from command-line, look it up with kpsewhich: */
   if (test_file('z', cnf_file)) {
-    cnf_file = kpse_find_file(cnf_default, kpse_web2c_format, false);
+    cnf_file = kpse_find_file(cnf_default, kpse_web2c_format, true);
     if (test_file('z', cnf_file)) {
       fprintf(stderr, "%s: fatal error, configuration file not found.\n",
 	      progname);
@@ -363,10 +425,13 @@ int main(int argc, char* argv[])
     exit(1);
   }
 
+  /* read the fmtutil.cnf file */
+  read_fmtutilcnf();
+
   if (cmd == SHOWHYPHEN) {
     exit(show_hyphen_file(arg));
   }
-  
+
   cache_vars();
 
   texmfmain = getval("MT_TEXMFMAIN");
@@ -377,10 +442,10 @@ int main(int argc, char* argv[])
 
   /* setup dest dir */
   if (test_file('z', destdir)) {
-    setval_default("TEXMF_CNF", expand_var("$TEXMF_CNF"));
-    if (test_file('z', getval("TEXMF_CNF")))
-      setval("TEXMF_CNF", expand_var("$MT_TEXMFMAIN"));
-    destdir = concat(getval("TEXMF_CNF"), "/web2c");
+    setval_default("VARTEXMF", expand_var("$VARTEXMF"));
+    if (test_file('z', getval("VARTEXMF")))
+      setval("VARTEXMF", expand_var("$MT_TEXMFMAIN"));
+    destdir = concat(getval("VARTEXMF"), "/web2c");
   }
   if (!test_file('d', destdir)) {
     fprintf(stderr, "%s: format directory `%s' does not exist.\n",
@@ -428,14 +493,16 @@ int main(int argc, char* argv[])
   /* export KPSE_DOT */
   xputenv("KPSE_DOT", cwd);
 
+#if 0
   texinputs = getval("TEXINPUTS");
   if (test_file('z', texinputs)) {
     xputenv("TEXINPUTS", concat(cwd, ENV_SEP_STRING));
   }
   else {
-    xputenv("TEXINPUTS", concat3(cwd, ENV_SEP_STRING, texinputs));
+    xputenv("TEXINPUTS", concatn(cwd, ENV_SEP_STRING, 
+				 texinputs, ENV_SEP_STRING, NULL));
   }
-
+#endif
   pushd(tmpdir);
   
   if ((fout = fopen(getval("STDOUT"), "w")) == NULL) {
@@ -457,9 +524,6 @@ int main(int argc, char* argv[])
 
   errstatus = 0;
 
-  /* read the fmtutil.cnf file */
-  read_fmtutilcnf();
-
   switch (cmd) {
   case ALL:
     errstatus = recreate_all();
@@ -474,11 +538,17 @@ int main(int argc, char* argv[])
     errstatus = recreate_by_hyphenfile(arg);
     break;
   }
-
+#if 0
+  __asm int 3;
+#endif
   /* Install the log files and format files */
   process_multiple_files("*.log", move_log);
   process_multiple_files("*.fmt", move_fmt);
-  
+  process_multiple_files("*.efmt", move_fmt);
+
+  if (dolinks)
+    link_formats();
+
   mt_exit(errstatus);
 
 }
@@ -509,84 +579,190 @@ void cache_vars()
   setval_default("MT_LSR_PATH", kpse_format_info[kpse_db_format].path);
 }
 
+void do_format(int f)
+{
+  string cmdfmt = "%s -ini %s=%s -progname=%s %s <nul";
+  string fmtswitch;
+
+  if (FILESTRCASEEQ(fmts[f].engine, "etex")
+      || FILESTRCASEEQ(fmts[f].engine, "pdfetex")) {
+    fmtswitch = "-efmt";
+  }
+  else {
+    fmtswitch = "-fmt";
+  }
+  if (sprintf(cmd_sys, cmdfmt, fmts[f].engine, fmtswitch, fmts[f].progname,
+	      fmts[f].progname, fmts[f].texargs) > 1023) {
+    fprintf(stderr, "Warning: do_format buffer overrun!\n");
+    fprintf(stderr, "no command run.\n");
+    return;
+  }
+  
+  /* run discarding stdout */
+  start_redirection(quiet);
+
+  fprintf(stderr, "Running: %s\n", cmd_sys);
+
+  if (!norebuild && system(cmd_sys) == -1) {
+    fprintf(stderr, "%s: command `%s' failed.\n", progname, cmd_sys);
+    unlink(fmts[f].format);
+  }
+
+  pop_fd();
+
+  if (dolinks) {
+    to_link[nb_links].src = fmts[f].engine;
+    to_link[nb_links].dst = fmts[f].progname;
+    nb_links++;
+  }
+}
+
 int recreate_all()
 {
   int i;
 
-  printf("creating all\n");
+  for (i = 0; i < nb_formats; i++) {
+    do_format(i);
+  }
 
   return 0;
 }
 
 int create_missing()
 {
-  printf("creating missing\n");
+  int i;
+  string destfmt;
+
+  for (i = 0; i < nb_formats; i++) {
+    destfmt = concat_pathes(destdir, fmts[i].format);
+    if (! test_file('f', destfmt)) {
+      do_format(i);
+    }
+    free(destfmt);
+  }
   return 0;
 }
 
 int create_one_format(string fmtname)
 {
-  printf("creating one format: %s\n", fmtname);
+  int i;
+
+  for (i = 0; i < nb_formats; i++) {
+    if (FILESTRCASEEQ(fmtname, fmts[i].progname)) {
+      do_format(i);
+      return 0;
+    }
+  }
+  fprintf(stderr, "%s: format %s not found in fmtutil.cnf.\n",
+	  progname, fmtname);
   return 0;
 }
 
 int recreate_by_hyphenfile(string hyphenfile)
 {
-  printf("recreating by hyphen file: %s\n", hyphenfile);
+  int i;
+
+  for (i = 0; i < nb_formats; i++) {
+    if (FILESTRCASEEQ(hyphenfile, fmts[i].hyphenation)) {
+      do_format(i);
+      return 0;
+    }
+  }
   return 0;
 }
 
-int show_hyphen_file(string hyphenfile)
+int show_hyphen_file(string format)
 {
-  printf("showhing hyphen file: %s\n", hyphenfile);
+  int i;
+  string hyphenfile;
+
+  for (i = 0; i < nb_formats; i++) {
+    if (FILESTRCASEEQ(format, fmts[i].progname)) {
+      if (STREQ(fmts[i].hyphenation, "-")) {
+	printf("%s\n", fmts[i].hyphenation);
+	return 0;
+      }
+      kpse_reset_program_name(format);
+      kpse_format_info[kpse_tex_format].cnf_path = kpse_cnf_get("TEXINPUTS");
+      if (!(hyphenfile 
+	    = kpse_find_file(fmts[i].hyphenation, kpse_tex_format, false))) {
+	fprintf(stderr, "%s: hyphen file %s not found\n", progname, 
+		fmts[i].hyphenation);
+	return -1;
+      }
+      printf("%s\n", hyphenfile);
+      free(hyphenfile);
+      return 0;
+    }
+  }
+  fprintf(stderr, "no info for format %s\n", format);
   return 0;
 }
 
+/*
+  This function expects a globbing file specification.
+  There is no recursive call here.
+*/
 void process_multiple_files(string filespec, process_fn fn)
 {
-    /* In depth traversal of the subdir tree */
+
+  char path[PATH_MAX];
+  int path_len;
 #if defined(_WIN32)
   WIN32_FIND_DATA find_file_data;
   HANDLE hnd;
 #else
-  DIR *dp;
-  struct dirent *ep;
+  glob_t gl;
 #endif
-  struct stat stat_buf;	/* We have to have one local because
-			   of after */
-  char path[PATH_MAX];
-  int path_len = strlen(path);
 
   strcpy(path, filespec);
 
-#if defined(_WIN32)
+  /* If it is a directory, then get all the files there */
+  if (test_file('d', path)) {
+    path_len = strlen(path);
     strcat(path, "/*");
-    hnd = FindFirstFile(path, &find_file_data);
-    while (hnd != INVALID_HANDLE_VALUE && 
-	   FindNextFile(hnd, &find_file_data) != FALSE) { 
-      if(!strcmp(find_file_data.cFileName, ".")
-	 || !strcmp(find_file_data.cFileName, "..")) 
+  }
+  else {
+    path_len = strlen(path) - strlen(xbasename(path)) -1;
+  }
+
+#if defined(_WIN32)
+  hnd = FindFirstFile(path, &find_file_data);
+  while (hnd != INVALID_HANDLE_VALUE) {
+    if(!strcmp(find_file_data.cFileName, ".")
+       || !strcmp(find_file_data.cFileName, "..")) 
 	continue;
-      path[path_len+1] = '\0';
-      strcat(path, find_file_data.cFileName);
-      (*fn)(path);
+    path[path_len+1] = '\0';
+    strcat(path, find_file_data.cFileName);
+    if (KPSE_DEBUG_P(FMTUTIL_DEBUG)) {
+      fprintf(stderr, "Processing %s\n", path);
     }
-    path[path_len] = '\0';
-    FindClose(hnd);
-#else
-    if ((dp = opendir(path))) {
-      while ((ep = readdir(dp)))
-	if (strcmp(ep->d_name, ".") &&
-	    strcmp(ep->d_name, "..")) {
-	  path[path_len] = '/';
-	  strcpy(path+path_len+1, ep->d_name);
-	  (*fn)(path);
-	}
-      path[path_len] = '\0';
-      closedir(dp);
-    }
-    else
-      perror(path);
+    (*fn)(path);
+      if (FindNextFile(hnd, &find_file_data) == FALSE)
+	break;
+  }
+  path[path_len+1] = '\0';
+  FindClose(hnd);
+
+#else /* ! WIN32 */
+
+  switch  (glob(path, GLOB_NOCHECK, NULL, &gl)) {
+  case 0:			/* success */
+      for (i = 0; i < gl.gl_pathc; i++) {
+	(*fn)(gl.gl_pathv[i]);
+      }
+      globfree(&gl);
+      break;
+
+  case GLOB_NOSPACE:
+    fprintf(stderr, "%s: Out of memory while globbing %s.\n",
+	    kpse_program_name, filespec);
+    mt_exit(1);
+    break;
+
+  default:
+    break;
+  }
 #endif
 
 }
@@ -606,6 +782,9 @@ void move_fmt(string fmtname)
   string texfmt, plainfmt;
   fmtdest = concat3(destdir, DIR_SEP_STRING, fmtname);
   unlink(fmtdest);
+  if (KPSE_DEBUG_P(FMTUTIL_DEBUG)) {
+    fprintf(stderr, "Moving %s to %s\n", fmtname, fmtdest);
+  }
   mvfile(fmtname, fmtdest);
   free(fmtdest);
   texfmt = concat3(destdir, DIR_SEP_STRING, "tex.fmt");
@@ -621,3 +800,77 @@ void move_fmt(string fmtname)
   }
   free(texfmt);
 }
+
+void link_formats()
+{
+  int i, j;
+  boolean skip;
+  string src, srcdll, srcexe, dstexe;
+
+  for (i = 0; i < nb_links; i++) {
+    src = expand_var(concat("$SELFAUTOLOC/", to_link[i].src));
+    /* Check that the destination is not among the ones that should
+     *never* be removed */
+    skip = false;
+    for (j = 0; unrmable_engines[j] != NULL; j++) {
+      if (FILESTRCASEEQ(to_link[i].dst, unrmable_engines[j])) {
+	fprintf(stderr, 
+		"%s: %s is a basic engine, won't remove it... skipping.\n",
+		kpse_program_name, unrmable_engines[j]);
+	skip = true;
+	break;
+      }
+    }
+    /* Any smarter way to do that ? */
+    if (skip)
+      break;
+
+    dstexe = expand_var(concatn("$SELFAUTOLOC/", to_link[i].dst, ".exe", NULL));
+    if (test_file('f', dstexe) && force == false) {
+      fprintf(stderr, "%s: destination %s already exists, skipping.\n", kpse_program_name, dstexe);
+      free(dstexe);
+      continue;
+    }
+    srcexe = concat(src, ".exe");
+#ifdef WIN32
+    srcdll = concat(src, ".dll");
+    if (FILESTRCASEEQ(srcexe, dstexe)) {
+      printf("same files: %s and %s, doing nothing\n", srcexe, dstexe);
+    }
+    else if (test_file('f', srcdll)) {
+      if (norebuild) {
+	printf("copy %s to %s\n", srcexe, dstexe);
+      }
+      else if (catfile(srcexe, dstexe, false) == false) {
+	fprintf(stderr, "%s: failed to copy %s to %s\n", 
+		kpse_program_name, srcexe, dstexe);
+      }
+    }
+    else {
+      sprintf(cmd_sys,"lnexe %s %s", srcexe, dstexe);
+      if (norebuild) {
+	printf("lnexe %s to %s\n", srcexe, dstexe);
+      }
+      else if (system(cmd_sys) == -1) {
+	fprintf(stderr, "%s: failed to lnexe %s to %s\n", 
+		kpse_program_name, srcexe, dstexe);
+      }
+    }
+    free(srcdll);
+#else
+    srcexe = concat(src, ".exe");
+    if (FILESTRCASEEQ(srcexe, dstexe)) {
+      printf("same files: %s and %s, doing nothing\n");
+    }
+    else if (norebuild) {
+      printf("copy %s to %s\n", srcexe, dstexe);
+    }
+    else if (catfile(srcexe, dstexe, false) == false) {
+      fprintf(stderr, "%s: failed to copy %s to %s\n", 
+	      kpse_program_name, srcexe, dstexe);
+    }
+#endif
+    free(srcexe);
+    free(dstexe);
+  }
+}
