@@ -7,7 +7,6 @@
 
 #include "dvips.h" /* The copyright notice there is included too! */
 #ifdef KPATHSEA
-#include <kpathsea/c-auto.h>
 #include <kpathsea/c-pathch.h>
 #include <kpathsea/proginit.h>
 #include <kpathsea/progname.h>
@@ -603,7 +602,7 @@ main P2C(int, argc, char **, argv)
         exit (0);
       } else if (strcmp (argv[1], "--version") == 0) {
         extern KPSEDLL char *kpathsea_version_string;
-        puts ("dvips(k) 5.85");
+        puts ("dvips(k) 5.86");
         puts (kpathsea_version_string);
         puts ("Copyright (C) 1999 Radical Eye Software.\n\
 There is NO warranty.  You may redistribute this software\n\
@@ -1114,7 +1113,7 @@ default:
    }
    if (oname[0] == '-' && oname[1] == 0)
       oname[0] = 0 ;
-   if (*oname == 0 && ! filter) {
+   else if (*oname == 0 && ! filter) {
       oname = nextstring ;
 #ifndef VMCMS  /* get stuff before LAST "." */
       lastext = strlen(iname) - 1 ;
